@@ -8,7 +8,4 @@ from .types import (
     Item,
     Response
 )
-from logging import getLogger
-
-logger = getLogger()
-logger.setLevel("DEBUG")
\ No newline at end of file
+from .router import logger
\ No newline at end of file
