@@ -1,25 +1,26 @@
-from setuptools import find_packages
 from importlib import import_module
 from inspect import getmembers
-from os import path
-from pkgutil import walk_packages
+from glob import glob
 from appsync_router import Router, logger
 
 logger.setLevel("DEBUG")
 
 
+def find_packages():
+    packages = []
+    for x in glob(f"{__name__}/**/*.py", recursive=True):
+        pkg_name = x.replace(".py", "").replace("/", ".")
+        if pkg_name != f"{__name__}.__init__":
+            packages.append(pkg_name)
+
+    return packages
+
+
 def get_route_handlers():
-    resolvers = []
-    for p in find_packages("."):
-        resolvers += [f"{p}.{x.name}" for x in walk_packages(
-            path=[p.replace(".", "/")])]
-
     package_list = {}
-
-    for package in resolvers:
+    for package in find_packages():
         pkg_name = package.split(".")[-1]
         pkg_root = ".".join(package.split(".")[:-1])
-
         mod = import_module(f".{pkg_name}", pkg_root)
         callables = []
         for _, item in getmembers(mod):
